import os
from abc import ABC
from os import path
from typing import Optional, Tuple, Union

import numpy as np
from dynesty import NestedSampler, DynamicNestedSampler

from autoconf import conf
from autofit import exc
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.nest.abstract_nest import AbstractNest
from autofit.non_linear.nest.dynesty.plotter import DynestyPlotter
from autofit.plot.output import Output


def prior_transform(cube, model):
    phys_cube = model.vector_from_unit_vector(
        unit_vector=cube, ignore_prior_limits=True
    )

    for i in range(len(phys_cube)):
        cube[i] = phys_cube[i]

    return cube


class AbstractDynesty(AbstractNest, ABC):
    def __init__(
        self,
        name: str = "",
        path_prefix: str = "",
        unique_tag: Optional[str] = None,
        prior_passer: PriorPasser = None,
        iterations_per_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs,
    ):
        """
        A Dynesty non-linear search.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        iterations_per_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs,
        )

        self.logger.debug(f"Creating {self.__class__.__name__} Search")

    class Fitness(AbstractNest.Fitness):
        @property
        def resample_figure_of_merit(self):
            """
            If a sample raises a FitException, this value is returned to signify that the point requires resampling or
            should be given a likelihood so low that it is discard.

            -np.inf is an invalid sample value for Dynesty, so we instead use a large negative number.
            """
            return -1.0e99

        def history_save(self):
            pass

    def _fit(
        self,
        model: AbstractPriorModel,
        analysis,
        log_likelihood_cap: Optional[float] = None,
    ):
        """
        Fit a model using Dynesty and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        By default, Dynesty runs using an in-built multiprocessing Pool option. This occurs even
        if `number_of_cores=1`, because the dynesty savestate includes this pool, meaning that a resumed run can
        then increase the `number_of_cores`.

        However, certain operating systems (e.g. Windows) do not support Python multiprocessing particularly well.
        This can cause Dynesty to crash when a pool is included. If this occurs (raising a `RunTimeException`)
        a Dynesty object without a pool is created and used instead.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data,
            returning the log likelihood dynesty maximizes.
        log_likelihood_cap
            An optional cap to the log likelihood values, which means all likelihood evaluations above this value
            are rounded down to it. This is used to remove numerical instability in an Astronomy based project.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted samples of the fit.
        """

        from dynesty.pool import Pool

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis, log_likelihood_cap=log_likelihood_cap,
        )

        if os.path.exists(self.checkpoint_file):
            self.logger.info(
                "Existing Dynesty samples found, resuming non-linear search."
            )
        else:
            self.logger.info(
                "No Dynesty samples found, beginning new non-linear search. "
            )

        finished = False

        while not finished:

            checkpoint_exists = os.path.exists(self.checkpoint_file)

            try:

                if conf.instance["non_linear"]["nest"][self.__class__.__name__][
                    "parallel"
                ].get("force_x1_cpu") or self.kwargs.get("force_x1_cpu"):

                    raise RuntimeError

                with Pool(
                    njobs=self.number_of_cores,
                    loglike=fitness_function,
                    prior_transform=prior_transform,
                    logl_args=(model, fitness_function),
                    ptform_args=(model,),
                ) as pool:

                    sampler = self.sampler_from(
                        model=model,
                        fitness_function=fitness_function,
                        checkpoint_exists=checkpoint_exists,
                        pool=pool,
                        queue_size=self.number_of_cores,
                    )

                    finished = self.run_sampler(sampler=sampler)

            except RuntimeError:

                checkpoint_exists = os.path.exists(self.checkpoint_file)

                if not checkpoint_exists:

                    self.logger.info(
                        """
                        Your operating system does not support Python multiprocessing.
                        
                        A single CPU non-multiprocessing Dynesty run is being performed.
                        """
                    )

                sampler = self.sampler_from(
                    model=model,
                    fitness_function=fitness_function,
                    checkpoint_exists=checkpoint_exists,
                    pool=None,
                    queue_size=None,
                )

                finished = self.run_sampler(sampler=sampler)

            self.perform_update(model=model, analysis=analysis, during_analysis=True)

    def iterations_from(
        self, sampler: Union[NestedSampler, DynamicNestedSampler]
    ) -> Tuple[int, int]:
        """
        Returns the next number of iterations that a dynesty call will use and the total number of iterations
        that have been performed so far.

        This is used so that the `iterations_per_update` input leads to on-the-fly output of dynesty results.

        It also ensures dynesty does not perform more samples than the `maxcall` input variable.

        Parameters
        ----------
        sampler
            The Dynesty sampler (static or dynamic) which is run and performs nested sampling.

        Returns
        -------
        The next number of iterations that a dynesty run sampling will perform and the total number of iterations
        it has performed so far.
        """
        try:
            total_iterations = np.sum(sampler.results.ncall)
        except AttributeError:
            total_iterations = 0

        if self.config_dict_run.get("maxcall") is not None:
            iterations = self.config_dict_run["maxcall"] - total_iterations

            return int(iterations), int(total_iterations)
        return self.iterations_per_update, int(total_iterations)

    def run_sampler(self, sampler: Union[NestedSampler, DynamicNestedSampler]):
        """
        Run the Dynesty sampler, which could be either the static of dynamic sampler.

        The number of iterations which the sampler runs from depends on the `maxcall` input. Due to on-to-fly updates
        via `perform_update` the number of remaining iterations compared to `maxcall` is tracked between every
        `run_nested` call, and whether or not the sampler is finished is returned at the end of this function.

        A second finish criteria is used, which occurs when the dynesty run performs zero likelihood updates (because
        the sampling accrording to Dynesty's termination criteria is complete).

        Parameters
        ----------
        sampler
            The Dynesty sampler (static or dynamic) which is run and performs nested sampling.

        Returns
        -------

        """

        iterations, total_iterations = self.iterations_from(sampler=sampler)

        config_dict_run = {key: value for key, value in self.config_dict_run.items() if key != 'maxcall'}

        if iterations > 0:
            sampler.run_nested(
                maxcall=iterations,
                print_progress=not self.silence,
                checkpoint_file=self.checkpoint_file,
                **config_dict_run,
            )

        iterations_after_run = np.sum(sampler.results.ncall)

        return (
            total_iterations == iterations_after_run
            or total_iterations == self.config_dict_run.get("maxcall")
        )

    def write_uses_pool(self, uses_pool: bool) -> str:
        """
        If a Dynesty fit does not use a parallel pool, and is then resumed using one,
        this causes significant slow down.

        This file checks the original pool use so an exception can be raised to avoid this.
        """
        with open(path.join(self.paths.samples_path, "uses_pool.save"), "w+") as f:
            if uses_pool:
                f.write("True")
            else:
                f.write("")

    def read_uses_pool(self) -> str:
        """
        If a Dynesty fit does not use a parallel pool, and is then resumed using one,
        this causes significant slow down.

        This file checks the original pool use so an exception can be raised to avoid this.
        """
        with open(path.join(self.paths.samples_path, "uses_pool.save"), "r+") as f:
            return bool(f.read())

    @property
    def checkpoint_file(self) -> str:
        """
        The path to the file used by dynesty for checkpointing.
        """
        return path.join(self.paths.samples_path, "savestate.save")

    def config_dict_with_test_mode_settings_from(self, config_dict):

        return {
            **config_dict,
            "maxiter": 1,
            "maxcall": 1,
        }

    def live_points_init_from(self, model, fitness_function):
        """
        By default, dynesty live points are generated via the sampler's in-built initialization.

        However, in test-mode this would take a long time to run, thus we overwrite the initial live points
        with quickly generated samplers from the initializer.

        Parameters
        ----------
        model
        fitness_function

        Returns
        -------

        """

        (
            unit_parameters,
            parameters,
            log_likelihood_list,
        ) = self.initializer.samples_from_model(
            total_points=self.total_live_points,
            model=model,
            fitness_function=fitness_function,
        )

        init_unit_parameters = np.zeros(
            shape=(self.total_live_points, model.prior_count)
        )
        init_parameters = np.zeros(
            shape=(self.total_live_points, model.prior_count)
        )
        init_log_likelihood_list = np.zeros(shape=(self.total_live_points))

        for i in range(len(parameters)):
            init_unit_parameters[i, :] = np.asarray(unit_parameters[i])
            init_parameters[i, :] = np.asarray(parameters[i])
            init_log_likelihood_list[i] = np.asarray(log_likelihood_list[i])

        live_points = [
            init_unit_parameters,
            init_parameters,
            init_log_likelihood_list,
        ]

        blobs = np.asarray(self.total_live_points * [False])

        live_points.append(blobs)

        return live_points

    def sampler_from(
        self,
        model: AbstractPriorModel,
        fitness_function,
        checkpoint_exists: bool,
        pool: Optional,
        queue_size: Optional[int],
    ):
        raise NotImplementedError()

    def check_pool(self, uses_pool: bool, pool):

        if (uses_pool and pool is None) or (not uses_pool and pool is not None):
            raise exc.SearchException(
                """
                A Dynesty sampler has been loaded and its pool type is not the same as the input pool type.

                This means that the original samples in dynesty were computed with or without a 
                multiprocessing pool, whereas the run is now trying to use a multiprocessing pool.

                This could indiciate the number of cores have change values or Python multiprocessing
                has been disabled and then enabled.
                """
            )

    def samples_from(self, model):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For Dynesty, all information that we need is available from the instance of the dynesty sampler.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """
        raise NotImplementedError()

    def remove_state_files(self):

        os.remove(self.checkpoint_file)

    @property
    def total_live_points(self):
        raise NotImplementedError()

    def plot_results(self, samples):

        if not samples.pdf_converged:
            return

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["dynesty"][name]

        plotter = DynestyPlotter(
            samples=samples,
            output=Output(
                path=path.join(self.paths.image_path, "search"), format="png"
            ),
        )

        if should_plot("cornerplot"):
            plotter.cornerplot()

        if should_plot("traceplot"):
            plotter.traceplot()

        # There is currently a bug internal in dynesty where the matplotlib figure produced after these plots
        # is not closed, and has weird extra stuff on. I have commented these out for now, in the hope that dynesty
        # fix this bug in the future.

        # if should_plot("runplot"):
        #     plotter.runplot()

        # if should_plot("cornerpoints"):
        #     plotter.cornerpoints()



from abc import ABC, abstractmethod
from itertools import count, chain
from typing import (
    List,
    Tuple,
    Dict,
    Set,
    Optional,
    Collection,
    Any,
    TYPE_CHECKING, 
)

import numpy as np

from autoconf import cached_property
from autofit.graphical.utils import (
    FlattenArrays,
    nested_filter,
    nested_update,
    nested_zip,
    nested_set, 
    nested_map, 
    nested_items, 
    is_variable,
    Status, 
)
from autofit.mapper.variable import (
    Variable,
    Plate,
    FactorValue,
    VariableData,
)

if TYPE_CHECKING:
    from autofit.graphical.mean_field import MeanField
    from autofit.graphical.expectation_propagation import EPMeanField

Protocol = ABC  # for python 3.7 compat

Value = Dict[Variable, np.ndarray]

GradientValue = VariableData
HessianValue = Any


class FactorInterface(Protocol):
    def __call__(self, values: Value) -> FactorValue:
        pass


class FactorGradientInterface(Protocol):
    def __call__(self, values: Value) -> Tuple[FactorValue, GradientValue]:
        pass


class AbstractNode(ABC):
    _plates: Tuple[Plate, ...] = ()
    _factor: callable = None
    label = None
    _id = count()
    eps = 1e-6

    def __init__(
            self,
            plates: Tuple[Plate, ...] = (),
            factor_out=FactorValue,
            name="",
            **kwargs: Variable,
    ):
        """
        A node in a factor graph

        Parameters
        ----------
        args
            Positional arguments passed to the factor
        kwargs
            Key word arguments passed to the value
        """
        self._plates = plates
        self._kwargs = kwargs
        self._name = name or type(self).__name__ + str(tuple(self.variables))
        self._factor_out = factor_out
        self.id = next(self._id)

    def resolve_variable_dict(
            self, values: Dict[Variable, np.ndarray]
    ) -> Dict[str, np.ndarray]:
        return {
            self.variable_name_kw[v.name]: x
            for v, x in values.items()
            if v.name in self.variable_name_kw
        }

    def resolve_args(
            self, values: Dict[Variable, np.ndarray]
    ) -> Tuple[np.ndarray, ...]:
        return nested_update(self.args, values)

    @cached_property
    def fixed_values(self) -> VariableData:
        return VariableData()

    @cached_property
    def variables(self) -> Set[Variable]:
        """
        Dictionary mapping the names of variables to those variables
        """
        return frozenset(self.flat_args)

    @property
    def free_variables(self) -> Set[Variable]:
        return self.variables - self.fixed_values.keys()

    @property
    def kwargs(self) -> Dict[str, Variable]:
        return self._kwargs

    @kwargs.setter
    def kwargs(self, kwargs):
        del self.variables
        del self.name_variable_dict
        del self.variable_name_dict
        self._kwargs = kwargs

    @property
    def args(self) -> Tuple[Any, ...]:
        return tuple(self.kwargs.values())

    @property
    def arg_names(self) -> Tuple[str, ...]:
        return tuple(self.kwargs)
    
    @property 
    def flat_args(self) -> Tuple[Variable, ...]:
        return tuple(x for x, in nested_zip(self._kwargs))

    @property
    def factor_out(self):
        return self._factor_out

    @factor_out.setter
    def factor_out(self, factor_out):
        del self.deterministic_variables
        self._factor_out = factor_out

    @property
    def n_args(self) -> int:
        return len(self.args)

    @cached_property
    def name_variable_dict(self) -> Dict[str, Variable]:
        return {variable.name: variable for variable in self.variables}

    @cached_property
    def variable_name_dict(self) -> Dict[str, Variable]:
        return {variable.name: kw for kw, variable in self._kwargs.items()}

    @cached_property
    def deterministic_variables(self) -> Set[Variable]:
        deterministic_variables = set(
            v for v, in nested_filter(is_variable, self.factor_out)
        )
        deterministic_variables.discard(FactorValue)
        return frozenset(deterministic_variables)

    @property
    def plates(self) -> Tuple[Plate, ...]:
        return self._plates

    @cached_property
    def sorted_plates(self) -> Tuple[Plate, ...]:
        """
        A tuple of the set of all plates in this graph, ordered by id
        """
        return tuple(sorted(set(
            plate
            for variable in self.all_variables
            for plate in variable.plates
        )))

    def __getitem__(self, item):
        try:
            return self._kwargs[item]
        except KeyError as e:
            for variable in self.all_variables:
                if variable.name == item:
                    return variable
            raise AttributeError(f"No attribute {item}") from e

    @property
    def name(self) -> str:
        """
        A name for this object
        """
        return self._name

    @property
    def call_signature(self) -> str:
        """
        The apparent signature of this object
        """
        call_str = ", ".join(map("{0}={0}".format, self.kwarg_names))
        call_sig = f"{self.name}({call_str})"
        return call_sig

    @property
    def kwarg_names(self) -> List[str]:
        """
        The names of the variables passed as keyword arguments
        """
        return [arg.name for arg in self._kwargs.values()]

    @property
    def all_variables(self) -> Set[Variable]:
        """
        A dictionary of variables associated with this node
        """
        return self.variables.union(self.deterministic_variables)

    @property
    def ndim(self) -> int:
        """
        The number of plates contained within this graph's variables

        That is, the total dimensions of those variables.
        """
        return len(self.sorted_plates)

    def _match_plates(self, plates: Collection[Plate]) -> np.ndarray:
        """
        Find indices plates from some factor in the collection of
        plates associated with this node.

        Parameters
        ----------
        plates
            Plates from some other node

        Returns
        -------
        An array of plate indices
        """
        return np.array([self.plates.index(p) for p in plates], dtype=int)

    @abstractmethod
    def __call__(self, **kwargs) -> FactorValue:
        pass

    @property
    def _unique_representation(self):
        return (
            self._factor,
            self.arg_names,
            self.flat_args,
            self.deterministic_variables,
        )

    def __hash__(self):
        return hash(self._unique_representation)

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self._unique_representation == other._unique_representation
        return False

    def _factor_args(self, *args):
        return self._factor(**dict(zip(self.arg_names, args)))

    def _numerical_factor_jacobian(
            self, *args, eps: Optional[float] = None, **kwargs
    ) -> Tuple[Any, Any]:
        """Calculates the dense numerical jacobian matrix with respect to
        the input arguments, broadly speaking, the following should return the
        same values (within numerical precision of the finite differences)

        factor._numerical_factor_jacobian(*args)

        factor._factor(*args), jax.jacobian(factor._factor, range(len(args)))(*args)
        """
        eps = eps or self.eps

        args = nested_map(lambda _, val: np.array(val, dtype=np.float64), self.args, args)

        raw_fval0 = self._factor_args(*args)
        fval0 = self._factor_value(raw_fval0).to_dict()

        jac = {
            v0: nested_map(
                lambda _, v: np.empty_like(val, shape=np.shape(val) + np.shape(v)),
                self.args, args
            ) for v0, val in fval0.items()
        }
        for ks, _, val in nested_items(self.args, args):
            with np.nditer(val, op_flags=["readwrite"], flags=["multi_index"]) as it:
                for x_i in it:
                    val[it.multi_index] += eps
                    fval1 = self._factor_value(self._factor_args(*args)).to_dict()
                    jac_v1_i = (fval1 - fval0) / eps
                    x_i -= eps
                    indexes = (Ellipsis,) + it.multi_index
                    for v0, jac_v0v_i in jac_v1_i.items():
                        key_path = (v0, *ks, indexes)
                        nested_set(jac, key_path, jac_v0v_i)
                        # jac[v0][i][indexes] = jac_v0v_i

        # This replicates the output of normal
        # jax.jacobian(self.factor, len(self.args))(*args)
        jac_out = nested_update(self.factor_out, jac)

        return raw_fval0, jac_out

    def numerical_func_jacobian(
            self, values: VariableData, **kwargs
    ) -> tuple:
        args = self.resolve_args(values)
        raw_fval, raw_jac = self._numerical_factor_jacobian(*args, **kwargs)
        fval = self._factor_value(raw_fval)
        jvp = self._jac_out_to_jvp(raw_jac, values=fval.to_dict().merge(values))
        return fval, jvp

    def jacobian(self, values: Dict[Variable, np.array], **kwargs):
        return self.func_jacobian(values, **kwargs)[1]

    def func_gradient(self, values: VariableData) -> Tuple[FactorValue, GradientValue]:
        fval, fjac = self.func_jacobian(values)
        return fval, fjac.grad()

    def flatten(self, param_shapes: FlattenArrays) -> "FlattenedNode":
        return FlattenedNode(self, param_shapes)


class AbstractFactor(AbstractNode, ABC):
    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name

    def __mul__(self, other):
        """
        When two factors are multiplied together this creates a graph
        """
        from autofit.graphical.factor_graphs.graph import FactorGraph

        return FactorGraph([self]) * other

    @property
    def _kwargs_dims(self) -> Dict[str, int]:
        """
        The number of plates for each keyword argument variable
        """
        return {key: len(value) for key, value in self._kwargs.items()}

    @cached_property
    def _variable_plates(self) -> Dict[str, np.ndarray]:
        """
        Maps the name of each variable to the indices of its plates
        within this node
        """
        return {
            variable: self._match_plates(variable.plates)
            for variable in self.all_variables
        }

    @property
    def n_deterministic(self) -> int:
        """
        How many deterministic variables are there associated with this node?
        """
        return len(self._deterministic_variables)

    def _set_factor(self, factor):
        self._factor = factor
        self._has_exact_projection = getattr(factor, "has_exact_projection", None)
        self._calc_exact_projection = getattr(factor, "calc_exact_projection", None)
        self._calc_exact_update = getattr(factor, "calc_exact_update", None)

    def resolve_args_and_out(self, values):
        if self.factor_out == FactorValue:
            return self.resolve_args(values)
        else:
            return chain(self.resolve_args(values), (nested_update(self.factor_out, values),))

    def has_exact_projection(self, mean_field) -> bool:
        if self._has_exact_projection:
            return self._has_exact_projection(*self.resolve_args_and_out(mean_field))
        return False

    def calc_exact_projection(self, mean_field) -> "MeanField":
        if self._calc_exact_projection:
            from autofit.graphical.mean_field import MeanField

            projection = self._calc_exact_projection(*self.resolve_args_and_out(mean_field))
            return MeanField(
                nested_filter(
                    is_variable, self.args + (self.factor_out,), projection
                )
            )
        else:
            raise NotImplementedError

    def calc_exact_update(self, mean_field) -> "MeanField":
        if self._calc_exact_update:
            from autofit.graphical.mean_field import MeanField

            projection = self._calc_exact_update(*self.resolve_args_and_out(mean_field))
            return MeanField(
                nested_filter(
                    is_variable, self.args + (self.factor_out,), projection
                )
            )
        else:
            raise NotImplementedError

    def name_for_variable(self, variable):
        return self.name

    @property
    def info(self):
        return repr(self)

    def __repr__(self) -> str:
        args = ", ".join(map(str, self.args))
        clsname = type(self).__name__
        if self.deterministic_variables:
            args += f", factor_out={self.factor_out}"

        return f"{clsname}({self.name}, {args})"

    def make_results_text(self, model_approx):
        """
        Create a string describing the posterior values after this factor
        during or after an EPOptimisation.
        Parameters
        ----------
        model_approx: EPMeanField
        Returns
        -------
        A string containing the name of this factor with the names and
        values of each associated variable in the mean field.
        """
        string = "\n".join(
            f"{variable} = {model_approx.mean_field[variable].mean}"
            for variable in self.variables
        )
        return f"{self.name}\n\n{string}"


class FlattenedNode:
    def __init__(self, node: "AbstractNode", param_shapes: FlattenArrays):
        self.node = node
        self.param_shapes = param_shapes

    def flatten(self, values: Value) -> np.ndarray:
        return self.param_shapes.flatten(values)

    def unflatten(self, x0: np.ndarray) -> Value:
        return self.param_shapes.unflatten(x0)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        values = self.unflatten(x)
        return self.node(values)

    def func_jacobian(self, x: np.ndarray):
        values = self.unflatten(x)
        fval, jval = self.node.func_jacobian(values)
        grad = self.flatten(jval)
        return fval, grad

    def func_jacobian_hessian(self, x: np.ndarray):
        values = self.unflatten(x)
        fval, jval, hval = self.node.func_jacobian_hessian(values)
        grad = self.flatten(jval)
        hess = self.param_shapes.flatten2d(hval)
        return fval, grad, hess

    def jacobian(self, x: np.ndarray):
        return self.func_jacobian(x)[1]

    def hessian(self, x: np.ndarray):
        return self.func_jacobian_hessian(x)[1]

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.node, name)



from typing import Optional

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings


class AbstractMCMC(NonLinearSearch):

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            initializer: Optional[Initializer] = None,
            auto_correlations_settings=AutoCorrelationsSettings(),
            iterations_per_update: Optional[int] = None,
            number_of_cores: Optional[int] = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        self.auto_correlations_settings = auto_correlations_settings
        self.auto_correlations_settings.update_via_config(
            config=self.config_type[self.__class__.__name__]["auto_correlations"]
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

    @property
    def config_type(self):
        return conf.instance["non_linear"]["mcmc"]



from abc import ABC, abstractmethod
from functools import wraps
from typing import Tuple, List

import numpy as np
from scipy._lib._util import _asarray_validated
from scipy.linalg import (
    cho_factor,
    solve_triangular,
    get_blas_funcs,
    block_diag,
    qr_update,
    qr,
)

from autoconf import cached_property


class LinearOperator(ABC):
    """Implements the functionality of a linear operator.

    All linear operators can be expressed as a tensor/matrix
    However for some it there may be more efficient representations

    e.g. The Diagonal Matrix (see `DiagonalMatrix`),
    Sherman-Morrison Matrix (see `ShermanMorrison`)
    or the vector-Jacobian product of a function see jax.jvp [0]

    The class also has the attributes, lshape, rshape, lsize, rsize to allow multidimensional tensors to be used,
    see `ShermanMorrison`, `MultiVecOuterProduct`, or
    `autofit.messages.transform.MultinomialLogitTransform`
    for examples of this use case

    If `M` is the dense matrix represenation of the LinearOperator then the
    actions of the methods can be represented by the appropriate matrix
    operation,

    Methods
    -------
    __mul__(x):
        M.dot(x)
        M·x

    __rmul__(x):
        x.dot(M)
        x·M

    __rtruediv__(x):
        x.dot(inv(M))
        x / M

    ldiv(x):
        inv(M).dot(x)
        M \ x

    log_det():
        log(det(M))
        log |M|

    quad(x):
        x.T.dot(M).dot(x)
        xᵀ·M·x

    invquad(x):
        x.T.dot(inv(M)).dot(x)
        xᵀ·M⁻¹·x

    [0] https://jax.readthedocs.io/en/latest/notebooks/autodiff_cookbook.html#vector-jacobian-product
    """

    _ldim = 1
    is_diagonal = False

    @abstractmethod
    def __mul__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def ldiv(self, x: np.ndarray) -> np.ndarray:
        pass

    def inv(self):
        return InverseOperator(self)

    def from_dense(self, M, shape=None, ldim=None):
        return self.operator.from_dense(M, shape, ldim)

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, ...]:
        pass

    def __len__(self) -> int:
        return self.lsize

    @property
    def lsize(self) -> int:
        return np.prod(self.lshape, dtype=int)

    @property
    def rsize(self) -> int:
        return np.prod(self.rshape, dtype=int)

    @property
    def size(self) -> int:
        return np.prod(self.shape, dtype=int)

    @property
    def ldim(self):
        return self._ldim

    @property
    def rdim(self):
        return self.ndim - self.ldim

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def lshape(self):
        return self.shape[: self.ldim]

    @property
    def rshape(self):
        return self.shape[self.ldim :]

    @cached_property
    @abstractmethod
    def log_det(self):
        pass

    def quad(self, M: np.ndarray) -> np.ndarray:
        return (M * self).T * self

    def invquad(self, M: np.ndarray) -> np.ndarray:
        return (M / self).T / self

    def to_operator(self):
        return self

    def transform_bounds(self, bounds: List[Tuple]) -> List[Tuple]:
        """
        Convenience method for transforming the bounds of an
        operation by this matrix.

        Parameters
        ----------
        bounds
            A list of tuples, each describing the lower and upper
            bound for a given dimension.

            There should be N bounds corresponding to N dimensions
            for this NxN matrix.

        Returns
        -------
        The bounds transformed according to this transformation.
        """
        lower, upper = zip(*bounds)
        lower = self * lower
        upper = self * upper
        return list(zip(lower, upper))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc in (np.multiply, np.matmul):
            return self.__rmul__(inputs[0])
        elif ufunc is np.divide:
            return self.__rtruediv__(inputs[0])
        else:
            return NotImplemented

    def matrixabs(self):
        raise NotADirectoryError()


class InverseOperator(LinearOperator):
    def __init__(self, operator):
        self.operator = operator

    def __mul__(self, x: np.ndarray) -> np.ndarray:
        return self.operator.ldiv(x)

    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        return x * self.operator

    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        return x / self.operator

    def ldiv(self, x: np.ndarray) -> np.ndarray:
        return self * x

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.operator.shape

    def __len__(self) -> int:
        return self.operator.lsize

    @property
    def lsize(self) -> int:
        return np.prod(self.lshape, dtype=int)

    @property
    def rsize(self) -> int:
        return np.prod(self.operator.rshape, dtype=int)

    @cached_property
    def log_det(self):
        return -self.operator.log_det

    def quad(self, M: np.ndarray) -> np.ndarray:
        return self.operator.invquad(M)

    def invquad(self, M: np.ndarray) -> np.ndarray:
        return self.operator.quad(M)

    def to_dense(self):
        M = self.operator.to_dense()
        M2 = np.reshape(M, (self.lsize, self.rsize))
        return np.linalg.inv(M2).reshape(self.operator.shape)

    def to_operator(self):
        return self.from_dense(self.to_dense(), self.shape, self.ldim)

    def diagonal(self):
        return self.to_operator().diagonal()

    def inv(self):
        return self.operator

    @property
    def is_diagonal(self):
        return self.operator.is_diagonal


class IdentityOperator(LinearOperator):
    is_diagonal = True

    def __init__(self):
        pass

    def _identity(self, values: np.ndarray) -> np.ndarray:
        return values

    __mul__ = _identity
    __rtruediv__ = _identity
    __rmul__ = _identity
    ldiv = _identity
    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    quad = _identity
    invquad = _identity

    @cached_property
    def log_det(self):
        return 0.0

    @property
    def shape(self):
        return ()

    def __len__(self):
        return 0


# TODO refactor these for non-square transformations


def _wrap_leftop(method):
    @wraps(method)
    def leftmethod(self, x: np.ndarray) -> np.ndarray:
        x = np.asanyarray(x)
        outshape = self.lshape + x.shape[self.rdim :] if self.ndim else x.shape
        return method(self, x.reshape(self.rsize, -1)).reshape(outshape)

    return leftmethod


def _wrap_rightop(method):
    @wraps(method)
    def rightmethod(self, x: np.ndarray) -> np.ndarray:
        x = np.asanyarray(x)
        outshape = x.shape[: -self.ldim] + self.rshape if self.ndim else x.shape
        return method(self, x.reshape(-1, self.lsize)).reshape(outshape)

    return rightmethod


class MatrixOperator(LinearOperator):
    def __init__(self, M: np.ndarray, shape=None, ldim=None):
        self._shape = shape or np.shape(M)
        self._ldim = len(self.shape) // 2 if ldim is None else ldim
        self._M = np.asanyarray(M).reshape(self.shape)
        self._M2D = self._M.reshape(self.lsize, self.rsize)

    def __getitem__(self, index):
        M = self.operator.to_dense()[index]
        return self.from_dense(M, ldim=self.ldim)

    def reshape(self, shape):
        return self.from_dense(self.operator.to_dense().reshape(shape), ldim=self.ldim)

    @classmethod
    def from_dense(
        cls, M: np.ndarray, shape: Tuple[int, ...] = None, ldim: int = None
    ) -> "MatrixOperator":
        return cls(M, shape, ldim)

    @classmethod
    def from_operator(cls, operator: "MatrixOperator") -> "MatrixOperator":
        return cls.from_dense(operator.to_dense(), operator.shape, operator.ldim)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @cached_property
    def log_det(self):
        return np.linalg.slogdet(self._M2D)[1]

    @_wrap_leftop
    def __mul__(self, x: np.ndarray) -> np.ndarray:
        return self._M2D.dot(x)

    @_wrap_rightop
    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        return np.linalg.solve(self._M2D.T, x.T)

    @_wrap_rightop
    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        return np.dot(x, self._M2D)

    @_wrap_leftop
    def ldiv(self, x: np.ndarray) -> np.ndarray:
        return np.linalg.solve(self._M2D, x)

    def to_dense(self):
        return self._M.copy()

    def diagonal(self) -> np.ndarray:
        return (
            self.to_dense()
            .reshape(self.rsize, self.lsize)
            .diagonal()
            .reshape(self.rshape)
        )

    def to_diagonal(self):
        return DiagonalMatrix(self.diagonal())

    def __add__(self, other: "MatrixOperator") -> "MatrixOperator":
        return type(self).from_dense(
            self.to_dense() + other.to_dense(), self.shape, self.ldim
        )

    def __sub__(self, other: "MatrixOperator") -> "MatrixOperator":
        return type(self).from_dense(
            self.to_dense() - other.to_dense(), self.shape, self.ldim
        )

    def update(self, *args):
        M = self.to_dense().reshape(self.lsize, self.rsize)
        for (u, v) in args:
            M += u.ravel()[:, None] * v.ravel()[None, :]

        return type(self).from_dense(M, self.shape, self.ldim)

    def lowrankupdate(self, *args):
        return self.update(*((u, u) for u in args))

    def lowrankdowndate(self, *args):
        return self.update(*((u, -u) for u in args))

    def diagonalupdate(self, d):
        M = self._M.copy()
        # set diagonal
        M.flat[:: self.lsize + 1] += np.asanyarray(d).ravel()
        return type(self).from_dense(M, self.shape, self.ldim)


def _mul_triangular(
    c, b, trans=False, lower=True, overwrite_b=False, check_finite=True
):
    """wrapper for BLAS function trmv to perform triangular matrix
    multiplications


    Parameters
    ----------
    a : (M, M) array_like
        A triangular matrix
    b : (M,) or (M, N) array_like
        vector/matrix being multiplied
    lower, optional
        Use only data contained in the lower triangle of `a`.
        Default is to use upper triangle.
    trans, optional
        type of multiplication,

        ========  =========
        trans     system
        ========  =========
        False     a b
        True      a^T b
    overwrite_b, optional
        Allow overwriting data in `b` (may enhance performance)
        not fully tested
    check_finite, optional
        Whether to check that the input matrices contain only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.
    """
    a1 = _asarray_validated(c, check_finite=check_finite)
    b1 = _asarray_validated(b, check_finite=check_finite)

    n = c.shape[1]
    if c.shape[0] != n:
        raise ValueError("Triangular matrix passed must be square")
    if b.shape[0] != n:
        raise ValueError(
            f"shapes {c.shape} and {b.shape} not aligned: "
            f"{n} (dim 1) != {b.shape[0]} (dim 0)"
        )

    (trmv,) = get_blas_funcs(("trmv",), (a1, b1))
    if a1.flags.f_contiguous:

        def _trmv(a1, b1, overwrite_x):
            return trmv(a1, b1, lower=lower, trans=trans, overwrite_x=overwrite_x)

    else:
        # transposed system is solved since trmv expects Fortran ordering
        def _trmv(a1, b1, overwrite_x=overwrite_b):
            return trmv(
                a1.T, b1, lower=not lower, trans=not trans, overwrite_x=overwrite_x
            )

    if b1.ndim == 1:
        return _trmv(a1, b1, overwrite_b)
    elif b1.ndim == 2:
        # trmv only works for vector multiplications
        # set Fortran order so memory contiguous
        b2 = np.array(b1, order="F")
        for i in range(b2.shape[1]):
            # overwrite results
            _trmv(a1, b2[:, i], True)

        if overwrite_b:
            b1[:] = b2
            return b1
        else:
            return b2
    else:
        raise ValueError("b must have 1 or 2 dimensions, has {b.ndim}")


class QROperator(MatrixOperator):
    """This stores the matrix in the QR decomposition form"""

    def __init__(self, Q, R, shape=None, ldim=None):
        self.Q = Q
        self.R = R
        self._shape = shape or self.Q.shape
        self._ldim = ldim or len(self._shape) // 2

    @cached_property
    def M(self) -> np.ndarray:
        return self.Q.dot(self.R)

    @classmethod
    def from_dense(cls, hess, shape=None, ldim=None):
        shape = np.shape(hess)
        ldim = ldim or np.ndim(hess) // 2
        length = np.prod(shape[:ldim])
        M = np.reshape(hess, (length, length))
        return cls(*qr(M), shape=shape, ldim=ldim)

    @_wrap_leftop
    def __mul__(self, x):
        return self.M.dot(x)

    @_wrap_rightop
    def __rmul__(self, x):
        return x @ self.M

    @_wrap_rightop
    def __rtruediv__(self, x):
        return solve_triangular(self.R.T, x.T @ self.Q.T, lower=True).T

    @_wrap_leftop
    def ldiv(self, x):
        return self.Q.T @ solve_triangular(self.R, x, lower=False)

    @cached_property
    def log_det(self):
        return np.sum(np.log(self.R.diagonal()))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    def to_dense(self):
        return self.M.copy()

    def update(self, *args):
        Q, R = self.Q, self.R
        for u, v in args:
            Q, R = qr_update(Q, R, np.ravel(u), np.ravel(v))
        return QROperator(Q, R, self.shape, self.ldim)


class CholeskyOperator(MatrixOperator):
    """This performs the whitening transforms for the passed
    cholesky factor of the Hessian/inverse covariance of the system.

    see https://en.wikipedia.org/wiki/Whitening_transformation

    >>> M = CholeskyOperator(linalg.cho_factor(hess))
    >>> y = M * x
    >>> f, df_dx = func_and_gradient(M.ldiv(y))
    >>> df_dy = df_df * M
    >>>
    """

    def __init__(self, cho_factor, shape=None, ldim=None):
        self.c, self.lower = self.cho_factor = cho_factor
        self.L = self.c if self.lower else self.c.T
        self.U = self.c.T if self.lower else self.c
        self._shape = shape or self.L.shape
        self._ldim = ldim or len(self._shape) // 2

    @classmethod
    def from_dense(cls, hess, shape=None, ldim=None):
        shape = np.shape(hess)
        ldim = ldim or np.ndim(hess) // 2
        length = np.prod(shape[:ldim])
        return cls(cho_factor(np.reshape(hess, (length, length))), shape, ldim)

    @_wrap_leftop
    def __mul__(self, x):
        return _mul_triangular(self.U, x, lower=False)

    @_wrap_rightop
    def __rmul__(self, x):
        return _mul_triangular(self.L, x.T, lower=True).T

    @_wrap_rightop
    def __rtruediv__(self, x):
        return solve_triangular(self.L, x.T, lower=True).T

    @_wrap_leftop
    def ldiv(self, x):
        return solve_triangular(self.U, x, lower=False)

    @cached_property
    def log_det(self):
        return np.sum(np.log(self.U.diagonal()))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    def to_dense(self):
        return np.tril(self.L) @ np.triu(self.U)

    def matrixabs(self):
        return self


# This class may no longer be necessary?
class InvCholeskyTransform(CholeskyOperator):
    """In the case where the covariance matrix is passed
    we perform the inverse operations
    """

    __mul__ = CholeskyOperator.__rtruediv__
    __rmul__ = CholeskyOperator.ldiv
    __rtruediv__ = CholeskyOperator.__mul__
    ldiv = CholeskyOperator.__rmul__

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @cached_property
    def log_det(self):
        return -np.sum(np.log(self.U.diagonal()))

    def to_dense(self):
        return solve_triangular(self.U, np.triul(self.L), lower=False)


class DiagonalMatrix(MatrixOperator):
    """
    Represents the DiagonalMatrix with diagonal `scale`

    M = np.diag(scale.ravel())

    """

    is_diagonal = True

    def __init__(self, scale, inv_scale=None):
        self.scale = np.asanyarray(scale)
        self._fscale = np.ravel(self.scale)

        # Lazily calculate inverse
        if inv_scale is not None:
            self._finv_scale = np.ravel(inv_scale)

        self._ldim = self.scale.ndim

    def __repr__(self):
        return repr(self.scale)

    def __getitem__(self, index):
        if index[: self.ldim] == index[self.ldim :]:
            return DiagonalMatrix(self.scale[index[: self.ldim]])
        else:
            raise NotImplementedError("Can't get diagonal for non 'square' operators")

    def reshape(self, shape):
        shape = shape[: len(shape) // 2]
        return DiagonalMatrix(self.scale.reshape(shape))

    @cached_property
    def _finv_scale(self):
        return 1 / self._fscale

    def inv(self):
        return DiagonalMatrix(
            np.reshape(self._finv_scale, self.scale.shape), self.scale
        )

    @classmethod
    def from_dense(cls, M: np.ndarray, shape=None, ldim=None) -> "MatrixOperator":
        return MatrixOperator(M, shape, ldim)

    @_wrap_leftop
    def __mul__(self, x):
        return self._fscale[:, None] * x

    @_wrap_rightop
    def __rmul__(self, x):
        return x * self._fscale

    @_wrap_rightop
    def __rtruediv__(self, x):
        return x * self._finv_scale

    @_wrap_leftop
    def ldiv(self, x):
        return self._finv_scale[:, None] * x

    @cached_property
    def log_det(self):
        return np.sum(np.log(self.scale))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @property
    def shape(self):
        return self.scale.shape * 2

    def to_dense(self):
        return np.diag(self._fscale).reshape(self.shape)

    def diagonal(self):
        return self.scale

    def __add__(self, other):
        if isinstance(other, DiagonalMatrix):
            return DiagonalMatrix(self.scale + other.scale)
        else:
            return MatrixOperator(
                self.to_dense() + other.to_dense(), self.shape, self.ldim
            )

    def __sub__(self, other):
        if isinstance(other, DiagonalMatrix):
            return DiagonalMatrix(self.scale - other.scale)
        else:
            return MatrixOperator(
                self.to_dense() - other.to_dense(), self.shape, self.ldim
            )

    def diagonalupdate(self, d):
        return type(self)(self.scale + d)

    def matrixabs(self):
        return type(self)(abs(self.scale))


class VecOuterProduct(LinearOperator):
    """
    represents the matrix vector outer product

    outer = vec[:, None] * vecT[None, :]
    """

    def __init__(self, vec, vecT=None):
        self.vec = np.asanyarray(vec)
        self.vecT = np.asanyarray(vec if vecT is None else vecT)
        self._fvec = np.ravel(self.vec)[:, None]
        self._fvecT = np.ravel(self.vecT)[None, :]

    @_wrap_leftop
    def __mul__(self, x):
        return self._fvec @ self._fvecT.dot(x)

    @_wrap_rightop
    def __rmul__(self, x):
        return x.dot(self._fvec) @ self._fvecT

    def __rtruediv__(self, x):
        raise NotImplementedError()

    def ldiv(self, x):
        raise NotImplementedError()

    @cached_property
    def log_det(self):
        return -np.inf

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    @property
    def shape(self):
        return (self.vec.size, self.vecT.size)

    @property
    def ldim(self):
        return len(self.vec.shape)

    def to_dense(self):
        return np.outer(self._fvec, self._fvecT).reshape(self.shape)


class MultiVecOuterProduct(LinearOperator):
    """
    represents the matrix vector outer product for stacked vectors,

    outer -> block_diag(*
        (v[:, None] * u[None, :] for v, u in zip(vec, vecT)
    )
    outer @ x -> np.vstack([
        v[:, None] * u[None, :] @ x for v, u in zip(vec, vecT)
    ])
    """

    def __init__(self, vec, vecT=None):
        self.vec = np.asanyarray(vec)
        self.vecT = np.asanyarray(vec if vecT is None else vecT)
        self.n, self.d = self.vec.shape
        self.nT, self.dT = self.vecT.shape

    @_wrap_leftop
    def __mul__(self, x):
        return np.einsum(
            "ij,ik,ikl -> ijl", self.vec, self.vecT, x.reshape(*self.lshape, -1)
        )

    @_wrap_rightop
    def __rmul__(self, x):
        return np.einsum(
            "ij,ik,lij -> lik", self.vec, self.vecT, x.reshape(-1, *self.rshape)
        )

    def __rtruediv__(self, x):
        raise NotImplementedError()

    def ldiv(self, x):
        raise NotImplementedError()

    @cached_property
    def log_det(self):
        return -np.inf

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    @property
    def shape(self):
        return self.vec.shape + self.vecT.shape

    @property
    def lshape(self):
        return self.shape[:2]

    @property
    def rshape(self):
        return self.shape[2:]

    def to_dense(self):
        return block_diag(*(self.vec[:, :, None] * self.vecT[:, None, :])).reshape(
            self.shape
        )


class ShermanMorrison(LinearOperator):
    """
    Represents the Sherman-Morrison low rank update,
    inv(A + vec @ vec.T) =
        inv(A) - inv(A) @ vec @ vec.T @ inv(A) / (1 + vec @ inv(A) @ vec/T)
    """

    def __init__(self, linear, vec):
        self.linear = linear
        if np.ndim(vec) == 2:
            self.outer = MultiVecOuterProduct(vec)
        elif np.ndim(vec) == 1:
            self.outer = VecOuterProduct(vec)
        else:
            raise ValueError("vec must be 1 or 2 dimensional")

        self.inv_scale = 1 + linear.quad(vec)
        self._ldim = self.linear.ldim

    @_wrap_leftop
    def __mul__(self, x):
        return self.linear * x + self.outer * x

    @_wrap_rightop
    def __rmul__(self, x):
        return x * self.linear + x * self.outer

    @_wrap_rightop
    def __rtruediv__(self, x):
        x1 = x / self.linear
        return x1 - ((x1 / self.inv_scale) * self.outer) / self.linear

    @_wrap_leftop
    def ldiv(self, x):
        x1 = self.linear.ldiv(x)
        return x1 - self.linear.ldiv(self.outer.dot(x1 / self.inv_scale))

    @cached_property
    def log_det(self):
        return self.linear.log_det + np.log(self.inv_scale)

    @property
    def shape(self):
        return self.linear.shape

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    def to_dense(self):
        dense_outer = self.outer.to_dense()
        return self.linear.to_dense().reshape(self.shape) + dense_outer



import copy
from abc import ABC, abstractmethod
from typing import Optional, Set

from autofit.database.query import condition as c
from autofit.database.query.condition import Table


class NotCondition:
    def __init__(self, condition: c.AbstractCondition):
        """
        Prepend the condition with a 'not'

        Parameters
        ----------
        condition
            Some condition such equality to a value
        """
        self._condition = condition

    def __str__(self):
        return f"not ({self._condition})"


class AbstractQuery(c.AbstractCondition, ABC):
    def __init__(
            self,
            condition: Optional[
                c.AbstractCondition
            ] = None
    ):
        """
        A query run to find Fit instances that match given
        criteria

        Parameters
        ----------
        condition
            An optional condition
        """
        self._condition = condition

    @property
    def condition(self):
        return self._condition

    @property
    @abstractmethod
    def fit_query(self) -> str:
        """
        A full query that can be executed against the database to obtain
        fit ids
        """

    def __str__(self):
        return self.fit_query

    @property
    def tables(self) -> Set[Table]:
        return {c.fit_table}

    def __invert__(self):
        """
        Take ~ of this object.

        The object is copied and its condition is prepended
        with a 'not'.
        """
        inverted = copy.deepcopy(self)
        inverted._condition = NotCondition(
            self._condition
        )
        return inverted



from scipy.stats import stats
from .abstract import AbstractInterpolator


class LinearInterpolator(AbstractInterpolator):
    """
    Assume all attributes have a linear relationship with time
    """

    @staticmethod
    def _interpolate(x, y, value):
        slope, intercept, r, p, std_err = stats.linregress(x, y)
        return slope * value + intercept



from typing import Optional

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Sample
from autofit.non_linear.samples.nest import SamplesNest



class SamplesUltraNest(SamplesNest):

    @classmethod
    def from_results_internal(
            cls,
            results_internal,
            model: AbstractPriorModel,
            number_live_points: int,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the resultsof a `NonLinearSearch` (e.g.
        as files on your hard-disk) and Python.

        To create a `Samples` object after an `UltraNest` model-fit the results must be converted from the
        native format used by `UltraNest` to lists of values, the format used by the **PyAutoFit** `Samples` objects.
        This classmethod performs this conversion before creating a DyenstySamples` object.

        Parameters
        ----------
        results_internal
            The `UltraNest` results in their native internal format from which the samples are computed.
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        number_live_points
            The number of live points used by the `UltraNest` search.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        """
        
        parameters = results_internal["weighted_samples"]["points"]
        log_likelihood_list = results_internal["weighted_samples"]["logl"]
        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector)) for vector in parameters
        ]
        weight_list = results_internal["weighted_samples"]["weights"]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return SamplesUltraNest(
            model=model,
            sample_list=sample_list,
            number_live_points=number_live_points,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
            results_internal=results_internal,
        )

    @property
    def number_live_points(self):
        return self._number_live_points

    @property
    def total_samples(self):
        return self.results_internal["ncall"]

    @property
    def log_evidence(self):
        return self.results_internal["logz"]



from abc import ABC, abstractmethod
from typing import Set, Optional

from autofit.database.model import get_class_path


class Table:
    def __init__(self, name: str, abbreviation: Optional[str] = None):
        """
        A table containing some type of object in the database.

        object is parent to all other tables.
        value contains any numeric values.
        string_value contains any string values.

        Parameters
        ----------
        name
            The name of the table
        """
        self.name = name
        self._abbreviation = abbreviation

    @property
    def abbreviation(self) -> str:
        """
        A one letter abbreviation used as an alias for this table
        """
        if self._abbreviation is not None:
            return self._abbreviation
        return "".join(part[0] for part in self.name.split("_"))

    def __str__(self):
        """
        Describes the table in the FROM or JOIN statement
        """
        return f"{self.name} AS {self.abbreviation}"

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name

    def __hash__(self):
        return hash(self.name)


object_table = Table("object")
value_table = Table("value")
string_value_table = Table("string_value")
fit_table = Table("fit")
info_table = Table("info", "info")
none_table = Table("none")


class AbstractCondition(ABC):
    @property
    @abstractmethod
    def tables(self) -> Set[Table]:
        """
        The set of tables this condition applies to
        """

    @abstractmethod
    def __str__(self):
        """
        The condition written as SQL
        """

    def __and__(self, other: "AbstractCondition"):
        """
        Combine this and another query with an AND statement.

        Simplification is applied so that the query will execute as fast as possible.
        """
        from .junction import And

        return And(self, other)

    def __or__(self, other: "AbstractCondition"):
        """
        Combine this and another query with an AND statement.

        Simplification is applied so that the query will execute as fast as possible.
        """
        from .junction import Or

        return Or(self, other)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __gt__(self, other):
        return str(self) > str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __repr__(self):
        return f"<{self.__class__.__name__} {str(self)}>"


class NoneCondition(AbstractCondition):
    @property
    def tables(self) -> Set[Table]:
        return {none_table}

    def __str__(self):
        return "1 = 1"


class AbstractValueCondition(AbstractCondition, ABC):
    def __init__(self, symbol: str, value):
        """
        A condition which compares the named column to a value

        Parameters
        ----------
        symbol
            =, <=, >=, < or >
        value
            A string, int or float
        """
        self.value = value
        self.symbol = symbol


class ValueCondition(AbstractValueCondition):
    @property
    def tables(self) -> Set[Table]:
        """
        Tables included in the condition string
        """
        return {value_table}

    def __str__(self) -> str:
        """
        The condition in SQL
        """
        return f"{value_table.abbreviation}.value {self.symbol} {self.value}"


class StringValueCondition(AbstractValueCondition):
    @property
    def tables(self):
        """
        Tables included in the condition string
        """
        return {string_value_table}

    def __str__(self):
        """
        The condition in SQL
        """
        return f"{string_value_table.abbreviation}.value {self.symbol} '{self.value}'"


class NameCondition(AbstractCondition):
    def __init__(self, name: str):
        """
        Condition used to match the name of an object. e.g. galaxy, lens, brightness.

        Parameters
        ----------
        name
            The name of an attribute in the model
        """
        self.name = name

    @property
    def tables(self):
        """
        Tables included in the condition string
        """
        return {object_table}

    def __str__(self):
        """
        The condition in SQL
        """
        return f"{object_table.abbreviation}.name = '{self.name}'"


class TypeCondition(AbstractCondition):
    def __init__(self, cls: type):
        """
        Condition used to match the type of an object. e.g. SersicLightProfile

        Parameters
        ----------
        cls
            The type
        """
        self.cls = cls

    @property
    def tables(self):
        """
        Tables included in the condition string
        """
        return {object_table}

    def __str__(self):
        """
        The condition in SQL
        """
        return f"{object_table.abbreviation}.class_path = '{self.class_path}'"

    @property
    def class_path(self) -> str:
        """
        The full import path of the type
        """
        return get_class_path(self.cls)


class AttributeCondition(AbstractCondition, ABC):
    def __init__(self, attribute, value):
        self.attribute = attribute
        self._value = value

    @property
    def tables(self) -> Set[Table]:
        return {fit_table}

    @abstractmethod
    def __str__(self):
        pass


class EqualityAttributeCondition(AttributeCondition):
    @property
    def value(self):
        if isinstance(self._value, str):
            return f"'{self._value}'"
        return self._value

    def __str__(self):
        if self._value is None:
            return f"{self.attribute} IS NULL"
        return f"{self.attribute} = {self.value}"


class ContainsAttributeCondition(AttributeCondition):
    def __str__(self):
        return f"{self.attribute} LIKE '%{self._value}%'"


class InAttributeCondition(AttributeCondition):
    def __str__(self):
        return f"'{self._value}' LIKE '%' || {self.attribute} || '%'"


class AttributeCondition(AbstractCondition):
    def __init__(self, attribute):
        self.attribute = attribute

    @property
    def tables(self) -> Set[Table]:
        return {fit_table}

    def __str__(self):
        return self.attribute



import os
from codecs import open
from os.path import abspath, dirname, join
from os import environ

from setuptools import find_packages, setup

this_dir = abspath(dirname(__file__))
with open(join(this_dir, "README.rst"), encoding="utf-8") as file:
    long_description = file.read()

with open(join(this_dir, "requirements.txt")) as f:
    requirements = f.read().split("\n")

version = environ.get("VERSION", "1.0.dev0")
requirements.extend([
    f'autoconf=={version}'
])


def config_packages(directory):
    paths = [directory.replace("/", ".")]
    for (path, directories, filenames) in os.walk(directory):
        for directory in directories:
            paths.append(f'{path}/{directory}'.replace("/", "."))
    return paths


setup(
    name="autofit",
    version=version,
    description="Classy Probabilistic Programming",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/rhayes777/PyAutoFit",
    author="James Nightingale and Richard Hayes",
    author_email="richard@rghsoftware.co.uk",
    include_package_data=True,
    license="MIT License",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires='>=3.7',
    keywords="cli",
    packages=find_packages(exclude=["docs", "test_autofit", "test_autofit*"]) + config_packages('autofit/config'),
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
)



from typing import Optional

from autoconf import cached_property
from autofit.database.sqlalchemy_ import sa

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.optimize.abstract_optimize import AbstractOptimizer
from autofit.non_linear.abstract_search import Analysis
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import AbstractInitializer
from autofit.non_linear.optimize.lbfgs.samples import SamplesLBFGS

import copy
from scipy import optimize
import numpy as np


class LBFGS(AbstractOptimizer):

    __identifier_fields__ = ()

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        prior_passer: Optional[PriorPasser] = None,
        initializer: Optional[AbstractInitializer] = None,
        iterations_per_update: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs
    ):
        """
        A L-BFGS scipy non-linear search.

        For a full description of the scipy L-BFGS method, checkout its documentation:

        https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html

        If you use `LBFGS` as part of a published work, please cite the package via scipy following the instructions
        under the *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores: int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating LBFGS Search")

    @cached_property
    def config_dict_options(self):

        config_dict = copy.deepcopy(self._class_config["options"])

        for key, value in config_dict.items():
            try:
                config_dict[key] = self.kwargs[key]
            except KeyError:
                pass

        return config_dict

    def _fit(
        self,
        model: AbstractPriorModel,
        analysis: Analysis,
        log_likelihood_cap: Optional[float] = None,
    ):
        """
        Fit a model using the scipy L-BFGS method and the Analysis class which contains the data and returns the log
        likelihood from instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space.
        analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data,
            returning the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """
        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        if self.paths.is_object("x0"):

            x0 = self.paths.load_object("x0")
            total_iterations = self.paths.load_object("total_iterations")

            self.logger.info(
                "Existing LBGFS samples found, resuming non-linear search."
            )

        else:

            (
                unit_parameter_lists,
                parameter_lists,
                log_posterior_list,
            ) = self.initializer.samples_from_model(
                total_points=1, model=model, fitness_function=fitness_function,
            )

            x0 = np.asarray(parameter_lists[0])

            total_iterations = 0

            self.logger.info(
                "No LBFGS samples found, beginning new non-linear search. "
            )

        maxiter = self.config_dict_options.get("maxiter", 1e8)

        while total_iterations < maxiter:

            iterations_remaining = maxiter - total_iterations

            iterations = min(self.iterations_per_update, iterations_remaining)

            if iterations > 0:

                config_dict_options = self.config_dict_options
                config_dict_options["maxiter"] = iterations

                lbfgs = optimize.minimize(
                    fun=fitness_function.__call__,
                    x0=x0,
                    method="L-BFGS-B",
                    options=config_dict_options,
                    **self.config_dict_search
                )

                total_iterations += lbfgs.nit

                self.paths.save_object("total_iterations", total_iterations)
                self.paths.save_object(
                    "log_posterior",
                    fitness_function.log_posterior_from(parameter_list=lbfgs.x),
                )
                self.paths.save_object("x0", lbfgs.x)

                self.perform_update(
                    model=model, analysis=analysis, during_analysis=True
                )

                x0 = lbfgs.x

                if lbfgs.nit < iterations:
                    return

        self.logger.info("L-BFGS sampling complete.")

    def samples_from(self, model: AbstractPriorModel):

        return SamplesLBFGS.from_results_internal(
            model=model,
            results_internal=self.paths.load_object("x0"),
            log_posterior_list=np.array([self.paths.load_object("log_posterior")]),
            total_iterations=self.paths.load_object("total_iterations"),
            time=self.timer.time,
        )



import json
import os
import shutil
from os import path
from typing import Optional
import logging

import dill

from autoconf import conf
from autofit.text import formatter
from autofit.tools.util import open_
from .abstract import AbstractPaths
from ..samples import load_from_table
from autofit.non_linear.samples.pdf import SamplesPDF

logger = logging.getLogger(__name__)


class DirectoryPaths(AbstractPaths):
    def save_named_instance(self, name: str, instance):
        """
        Save an instance, such as that at a given sigma
        """
        self.save_object(name, instance)

    def _path_for_pickle(self, name: str):
        return path.join(self._pickle_path, f"{name}.pickle")

    def save_object(self, name: str, obj: object):
        """
        Serialise an object using dill and save it to the pickles
        directory of the search.

        Parameters
        ----------
        name
            The name of the object
        obj
            A serialisable object
        """
        with open_(self._path_for_pickle(name), "wb") as f:
            dill.dump(obj, f)

    def load_object(self, name: str):
        """
        Load a serialised object with the given name.

        e.g. if the name is 'model' then pickles/model.pickle is loaded.

        Parameters
        ----------
        name
            The name of a serialised object

        Returns
        -------
        The deserialised object
        """
        with open_(self._path_for_pickle(name), "rb") as f:
            return dill.load(f)

    def remove_object(self, name: str):
        """
        Remove the object with the given name from the pickles folder.

        Parameters
        ----------
        name
            The name of a pickle file excluding .pickle
        """
        try:
            os.remove(self._path_for_pickle(name))
        except FileNotFoundError:
            pass

    def is_object(self, name: str) -> bool:
        """
        Is there a file pickles/{name}.pickle?
        """
        return os.path.exists(self._path_for_pickle(name))

    @property
    def is_complete(self) -> bool:
        """
        Has the search been completed?
        """
        return path.exists(self._has_completed_path)

    def completed(self):
        """
        Mark the search as complete by saving a file
        """
        open_(self._has_completed_path, "w+").close()

    def load_samples(self):
        return load_from_table(filename=self._samples_file)

    def save_samples(self, samples):
        pass

    def samples_to_csv(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """
        if conf.instance["general"]["output"]["samples_to_csv"]:
            samples.write_table(filename=self._samples_file)
            samples.info_to_json(filename=self._info_file)
            if isinstance(samples, SamplesPDF):
                try:
                    samples.save_covariance_matrix(self._covariance_file)
                except ValueError as e:
                    logger.warning(
                        f"Could not save covariance matrix because of the following error:\n{e}"
                    )

    def load_samples_info(self):
        with open_(self._info_file) as infile:
            return json.load(infile)

    def save_all(self, search_config_dict=None, info=None, pickle_files=None):
        search_config_dict = search_config_dict or {}
        info = info or {}
        pickle_files = pickle_files or []

        self.save_identifier()
        self.save_parent_identifier()
        self._save_search(config_dict=search_config_dict)
        self._save_model_info(model=self.model)
        self._save_parameter_names_file(model=self.model)
        self.save_object("info", info)
        self.save_object("search", self.search)
        self.save_object("model", self.model)
        self._save_metadata(search_name=type(self.search).__name__.lower())
        self._move_pickle_files(pickle_files=pickle_files)

    @AbstractPaths.parent.setter
    def parent(self, parent: AbstractPaths):
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.
        """
        self._parent = parent

    def save_parent_identifier(self):
        if self.parent is not None:
            with open_(self._parent_identifier_path, "w+") as f:
                f.write(self.parent.identifier)
            self.parent.save_unique_tag()

    def save_unique_tag(self, is_grid_search=False):
        if is_grid_search:
            with open_(self._grid_search_path, "w+") as f:
                if self.unique_tag is not None:
                    f.write(self.unique_tag)

    @property
    def _parent_identifier_path(self) -> str:
        return path.join(self.output_path, ".parent_identifier")

    @property
    def _grid_search_path(self) -> str:
        return path.join(self.output_path, ".is_grid_search")

    @property
    def is_grid_search(self) -> bool:
        """
        Is this a grid search which comprises a number of child searches?
        """
        return os.path.exists(self._grid_search_path)

    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "AbstractPaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory.

        Parameters
        ----------
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.
        identifier

        Returns
        -------
        A new paths object
        """
        child = type(self)(
            name=name or self.name,
            path_prefix=path_prefix or self.path_prefix,
            is_identifier_in_paths=(
                is_identifier_in_paths
                if is_identifier_in_paths is not None
                else self.is_identifier_in_paths
            ),
            parent=self,
        )
        child.model = self.model
        child.search = self.search
        child._identifier = identifier
        return child

    def for_sub_analysis(self, analysis_name: str):
        """
        Paths for an analysis which is a child of another analysis.

        The analysis name forms a new directory on the end of the original
        analysis output path.
        """
        from .sub_directory_paths import SubDirectoryPaths

        return SubDirectoryPaths(parent=self, analysis_name=analysis_name)

    @property
    def _pickle_path(self) -> str:
        """
        This is private for a reason, use the save_object etc. methods to save and load pickles
        """
        return path.join(self.output_path, "pickles")

    def _save_metadata(self, search_name):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open_(path.join(self.output_path, "metadata"), "a") as f:
            f.write(
                f"""name={self.name}
            non_linear_search={search_name}
            """
            )

    def _move_pickle_files(self, pickle_files):
        """
        Move extra files a user has input the full path + filename of from the location specified to the
        pickles folder of the Aggregator, so that they can be accessed via the aggregator.
        """
        os.makedirs(self._pickle_path, exist_ok=True)
        if pickle_files is not None:
            [shutil.copy(file, self._pickle_path) for file in pickle_files]

    def _save_model_info(self, model):
        """
        Save the model.info file, which summarizes every parameter and prior.
        """
        with open_(path.join(self.output_path, "model.info"), "w+") as f:
            f.write(model.info)

    def _save_parameter_names_file(self, model):
        """
        Create the param_names file listing every parameter's label and Latex tag, which is used for corner.py
        visualization.

        The parameter labels are determined using the label.ini and label_format.ini config files.
        """

        parameter_names = model.model_component_and_parameter_names
        parameter_labels = model.parameter_labels
        subscripts = model.superscripts_overwrite_via_config
        parameter_labels_with_subscript = [
            f"{label}_{subscript}"
            for label, subscript in zip(parameter_labels, subscripts)
        ]

        parameter_name_and_label = []

        for i in range(model.prior_count):
            line = formatter.add_whitespace(
                str0=parameter_names[i],
                str1=parameter_labels_with_subscript[i],
                whitespace=70,
            )
            parameter_name_and_label += [f"{line}\n"]

        formatter.output_list_of_strings_to_file(
            file=path.join(self.samples_path, "model.paramnames"),
            list_of_strings=parameter_name_and_label,
        )

    @property
    def _info_file(self) -> str:
        return path.join(self.samples_path, "info.json")

    @property
    def _has_completed_path(self) -> str:
        """
        A file indicating that a `NonLinearSearch` has been completed previously
        """
        return path.join(self.output_path, ".completed")

    def _make_path(self) -> str:
        """
        Returns the path to the folder at which the metadata should be saved

        The path terminates with the identifier, unless the identifier has already
        been added to the path.
        """
        path_ = path.join(conf.instance.output_path, self.path_prefix, self.name)
        if self.is_identifier_in_paths:
            path_ = path.join(path_, self.identifier)
        return path_



#!/usr/bin/env python
"""
Usage
./add_notebook_quotes.py /path/to/input /path/to/output
"""

from sys import argv

from autofit.tools.add_notebook_quotes import add_notebook_quotes

if __name__ == "__main__":
    _, in_filename, out_filename = argv

    with open(in_filename) as f:
        lines = f.readlines()

    with open(out_filename, "w+") as f:
        f.writelines(
            add_notebook_quotes(
                lines
            )
        )


import logging
import warnings
from collections import abc
from enum import Enum
from functools import reduce
from operator import mul
from typing import Iterable, Tuple, TypeVar, Dict, NamedTuple, Optional, Union

import numpy as np
import six
from scipy.linalg import block_diag
from scipy.optimize import OptimizeResult

from autofit.mapper.variable import Variable, VariableData
from autofit.non_linear.result import Result


def try_getitem(value, index, default=None):
    try:
        return value[index]
    except TypeError:
        return default


class LogWarnings(warnings.catch_warnings):
    def __init__(self, *, module=None, messages=None, action=None, logger=logging.warning):
        super().__init__(record=True, module=module)
        self.messages = [] if messages is None else messages
        self.log = []
        self.action = action
        self.logger = logger

    def log_warning(self, warn):
        self.log.append(warn)
        warn_message = f"{warn.filename}:{warn.lineno}: {warn.message}"
        self.messages.append(warn_message)
        self.logger(warn_message)

    def __enter__(self):
        self.log = super().__enter__()
        self._module._showwarnmsg_impl = self.log_warning
        if self.action:
            warnings.simplefilter(self.action)

        return self


def is_variable(v, *args):
    return isinstance(v, Variable)


def is_iterable(arg):
    return isinstance(arg, abc.Iterable) and not isinstance(
        arg, six.string_types
    )

def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr(obj, '_fields')

def nested_getitem(obj, key):
    """
    Example
    -------
    >>> nested_getitem([1, (2, 3), [3, {'a': 1, 'b': 2}]], (2, 1, 'b'))
    2
    """
    for k in key:
        obj = obj[k]
    return obj 


def nested_get(obj, key, default=None):
    """
    Example
    -------
    >>> nested_get([1, (2, 3), [3, {'a': 1, 'b': 2}]], (2, 1, 'b'))
    2
    >>> nested_get([1, (2, 3), [3, {'a': 1, 'b': 2}]], (2, 1, 'c'), 'default')
    'default'
    """
    try:
        return nested_getitem(obj, key)
    except (KeyError, IndexError):
        return default
    

def nested_set(obj, key, val):
    """
    Example
    -------
    >>> obj = [1, (2, 3), [3, {'a': 1, 'b': 2}]]
    >>> nested_set(obj, (2, 1, 'b'), 3)
    >>> obj
    [1, (2, 3), [3, {'a': 1, 'b': 3}]]

    >>> nested_set(obj, (2, 1,), ()))
    >>> obj
    [1, (2, 3), [3, ()]]
    """
    *parents, k = key 
    for p in parents:
        obj = obj[p]
    obj[k] = val 


def nested_items(*args, key=(), key_func=None):
    """
    Example
    -------
    >>> list(nested_items([1, (2, 3), [3, {'a': 1, 'b': 2}]]))
    [((0,), 1), ((1, 0), 2), ((1, 1), 3), ((2, 0), 3), ((2, 1, 'a'), 1), ((2, 1, 'b'), 2)]

    >>> list(nested_items([1, (2, 3)], [4, (5, 6)]))
    [((0,), 1, 4), ((1, 0), 2, 5), ((1, 1), 3, 6)]
    """
    out, *_ = args
    if isinstance(out, dict):
        for k in sorted(out, key=key_func):
            yield from nested_items(*(out[k] for out in args), key=key + (k,), key_func=key_func)
    elif isinstance(out, (tuple, list)):
        for i, elems in enumerate(zip(*args)):
            yield from nested_items(*elems, key=key + (i,), key_func=key_func)
    else:
        yield (key,) + args


def nested_zip(*args, key_func=None):
    """ Iterates through a potentially nested set of list, tuples and dictionaries, 
    recursively looping through the structure and returning the leaves of the tree

    Example
    -------
    >>> list(nested_zip([1, (2, 3), [3, 2, {1, 2}]]))
    [(1,), (2,), (3,), (3,), (2,), (1,), (2,)]

    >>> list(nested_zip(
    ...     [1, (2, 3), [3, 2, {1, 2}]],
    ...     [1, ('a', 3), [3, 'b', {1, 'c'}]]
    ... ))
    [(1, 1), (2, 'a'), (3, 3), (3, 3), (2, 'b'), (1, 1), (2, 'c')]
    """
    out, *_ = args
    if isinstance(out, dict):
        for k in sorted(out, key=key_func):
            yield from nested_zip(*(out[k] for out in args), key_func=key_func)
    elif is_iterable(out):
        for elems in zip(*args):
            yield from nested_zip(*elems, key_func=key_func)
    else:
        yield args


def nested_iter(args, key_func=None):
    """Iterates through a potentially nested set of list, tuples and dictionaries, 
    recursively looping through the structure and returning the leaves of the tree

    Example
    -------
    >>> list(nested_iter([1, (2, 3), [3, 2, {1, 2}]]))    
    [1, 2, 3, 3, 2, 1, 2]
    """
    for elem, in nested_zip(args, key_func=key_func):
        yield elem


def nested_filter(func, *args, key_func=None):
    """ Iterates through a potentially nested set of list, tuples and dictionaries, 
    recursively looping through the structure and returning the arguments
    that func return true on, 

    Example
    -------
    >>> list(nested_filter(
    ...     lambda x, *args: x==2,
    ...     [1, (2, 3), [3, 2, {1, 2}]]
    ... ))
    [(2,), (2,), (2,)]

    >>> list(nested_filter(
    ...     lambda x, *args: x==2,
    ...     [1, (2, 3), [3, 2, {1, 2}]],
    ...     [1, ('a', 3), [3, 'b', {1, 'c'}]]
    ... ))
    [(2, 'a'), (2, 'b'), (2, 'c')]
    """
    for leaves in nested_zip(*args, key_func=key_func):
        if func(*leaves):
            yield leaves


def nested_map(func, *args, key_func=None):
    """
    Given a potentially nested set of list, tuples and dictionaries, recursively loop through the structure and
    replace any values that appear in the dict to_replace
    can set to replace dictionary keys optionally,

    Example
    -------
    >>> nested_map(lambda x: x*2, [1, (2, 3), [3, 2, {1, 2}, {'a': 'b'}]])
    [2, (4, 6), [6, 4, {2, 4}, {'a': 'bb'}]]

    >>> graph.utils.nested_map(
    ...     lambda x, y: x*y, 
    ...     [1, (2, 3), [3, {'a': 1, 'b': 2}]],
    ...     [1, (2, 3), [3, {'a': 1, 'b': 2}]]
    ... )
    [1, (4, 9), [9, {'a': 1, 'b': 4}]]
    """
    out, *_ = args
    if isinstance(out, dict):
        return type(out)(
            {
                k: nested_map(func, *(arg[k] for arg in args), key_func=key_func)
                for k in sorted(out, key=key_func)
            }
        )
    elif is_namedtuple(out):
        return type(out)(*(nested_map(func, *elems) for elems in zip(*args)))
    elif is_iterable(out):
        return type(out)(nested_map(func, *elems) for elems in zip(*args))

    return func(*args)


def nested_update(out, to_replace):
    """
    Given a potentially nested set of list, tuples and dictionaries, recursively loop through the structure and
    replace any values that appear in the dict to_replace
    can set to replace dictionary keys optionally,

    Does not replace values in place, so can 'mutate' tuples

    Example
    -------
    >>> nested_update([1, (2, 3), [3, 2, {1, 2}]], {2: 'a'})
    [1, ('a', 3), [3, 'a', {1, 'a'}]]

    >>> nested_update([{2: 2}], {2: 'a'})
    [{2: 'a'}]
    """

    def replace(val):
        return to_replace.get(val, val)
    
    return nested_map(replace, out)

from_variabledata = nested_update

def to_variabledata(variables, raw_values) -> VariableData:
    return VariableData(nested_filter(is_variable, variables, raw_values))


class StatusFlag(Enum):
    FAILURE = 0
    SUCCESS = 1
    NO_CHANGE = 2
    BAD_PROJECTION = 3

    @classmethod
    def get_flag(cls, success, n_iter):
        if success:
            if n_iter > 0:
                return cls.SUCCESS
            else:
                return cls.NO_CHANGE

        return cls.FAILURE


class Status:
    success: bool = True
    messages: Tuple[str, ...] = ()
    updated: bool = True
    flag: StatusFlag = StatusFlag.SUCCESS
    result: Optional[Result] = None

    def __init__(
            self,
            success: bool = True,
            messages: Tuple[str, ...] = (),
            updated: bool = True,
            flag: StatusFlag = StatusFlag.SUCCESS,
            result: Optional[Result] = None,
    ):
        self.success = success
        self.messages = messages
        self.updated = updated
        self.flag = flag
        self.result = result

    def __iter__(self):
        return iter((self.success, self.messages, self.updated, self.flag))

    def __bool__(self):
        return self.success

    def __str__(self):
        if self.success:
            return "Optimisation succeeded"
        return f"Optimisation failed: {self.messages}"

    def __repr__(self):
        return str(self)

    def _asdict(self):
        return dict(
            success=self.success,
            messages=self.messages,
            updated=self.updated,
            flag=self.flag,
        )


class FlattenArrays(dict):
    """
    >>> shapes = FlattenArrays(a=(1, 2), b=(2, 3))
    >>> shapes
    FlattenArrays(a=(1, 2), b=(2, 3))
    >>> shapes.flatten(
        a = np.arange(2).reshape(1, 2),
        b = np.arange(6).reshape(2, 3)**2)
    array([ 0,  1,  0,  1,  4,  9, 16, 25])
    >>> shapes.unflatten(
        [ 0,  1,  0,  1,  4,  9, 16, 25])
    {'a': array([[0, 1]]), 'b': array([[ 0,  1,  4],
        [ 9, 16, 25]])}
    """

    def __init__(self, dict_: Dict[Variable, Tuple[int, ...]]):
        super().__init__()

        self.update(dict_)
        self.splits = np.cumsum([np.prod(s) for s in self.values()], dtype=int)
        self.inds = [
            slice(i0, i1)
            for i0, i1 in
            # np.arange(i0, i1, dtype=int) for i0, i1 in
            zip(np.r_[0, self.splits[:-1]], self.splits)
        ]
        self.sizes = {k: np.prod(s, dtype=int) for k, s in self.items()}
        self.k_inds = dict(zip(self, self.inds))

    @classmethod
    def from_arrays(cls, arrays: Dict[str, np.ndarray]) -> "FlattenArrays":
        return cls({k: np.shape(arr) for k, arr in arrays.items()})

    def flatten(self, arrays_dict: Dict[Variable, np.ndarray]) -> np.ndarray:
        assert all(np.shape(arrays_dict[k]) == shape for k, shape in self.items())
        return np.concatenate([np.ravel(arrays_dict[k]) for k in self.keys()])

    def extract(self, key, flat, ndim=None):
        if ndim is None:
            ndim = len(flat.shape)

        ind = self.k_inds[key]
        return flat[(ind,) * ndim]

    def unflatten(self, arr: np.ndarray, ndim=None) -> Dict[str, np.ndarray]:
        arr = np.asanyarray(arr)
        if ndim is None:
            ndim = arr.ndim
        arrays = [arr[(ind,) * ndim] for ind in self.inds]
        arr_shapes = [arr.shape[ndim:] for arr in arrays]
        return VariableData({
            k: arr.reshape(shape * ndim + arr_shape)
            if shape or arr_shape
            else arr.item()
            for (k, shape), arr_shape, arr in zip(self.items(), arr_shapes, arrays)
        })

    def flatten2d(self, values: Dict[Variable, np.ndarray]) -> np.ndarray:
        assert all(np.shape(values[k]) == shape * 2 for k, shape in self.items())

        return block_diag(
            *(np.reshape(values[k], (n, n)) for k, n in self.sizes.items())
        )

    unflatten2d = unflatten

    def __repr__(self):
        shapes = ", ".join(map("{0[0]}={0[1]}".format, self.items()))
        return f"{type(self).__name__}({shapes})"

    @property
    def size(self):
        return self.splits[-1]


class OptResult(NamedTuple):
    mode: Dict[Variable, np.ndarray]
    hess_inv: Dict[Variable, np.ndarray]
    log_norm: float
    full_hess_inv: np.ndarray
    result: OptimizeResult
    status: Status = Status()


def gen_subsets(n, x, n_iters=None, rng=None):
    """
    Generates random subsets of length n of the array x, if the elements of
    x are unique then each subset will not contain repeated elements. Each 
    element is guaranteed to reappear after at most 2*len(x) new elements. 

    If `x` is a multi-dimensional array, it is only shuffled along its
first index.

    if x is an integer, generate subsets of ``np.arange(x)``.

    generates n_iters subsets before stopping. If n_iters is None then
    generates random subsets for ever

    rng is an optionally passed random number generator

    Examples
    --------
    >>> list(gen_subsets(3, 5, n_iters=3))
    [array([0, 2, 3]), array([1, 4, 0]), array([2, 3, 4])]
    >>> list(gen_subsets(3, [1,10,5,3], n_iters=3))
    [array([ 5, 10,  1]), array([3, 5, 1]), array([10,  3,  5])]
    """
    rng = rng or np.random.default_rng()
    x_shuffled = rng.permutation(x)
    tot = len(x_shuffled)

    i = 0
    stop = tot - n + 1
    iters = iter(int, 1) if n_iters is None else range(n_iters)
    for j in iters:
        if i < stop:
            yield x_shuffled[i: i + n]
            i += n
        else:
            x_shuffled = np.r_[x_shuffled[i:], rng.permutation(x_shuffled[:i])]
            yield x_shuffled[:n]
            i = n


def gen_dict(dict_gen):
    """
    Examples
    --------
    >>> list(gen_dict({1: gen_subsets(3, 4, 3), 2: gen_subsets(2, 5, 3)}))
    [{1: array([2, 1, 3]), 2: array([2, 0])},
     {1: array([0, 3, 1]), 2: array([3, 1])},
     {1: array([2, 0, 1]), 2: array([4, 2])}]
    """
    keys = tuple(dict_gen.keys())
    for val in zip(*dict_gen.values()):
        yield dict(zip(keys, val))


_M = TypeVar("_M")


def prod(iterable: Iterable[_M], *arg: Tuple[_M]) -> _M:
    """calculates the product of the passed iterable,
    much like sum, if a second argument is passed,
    this is the initial value of the calculation

    Examples
    --------
    >>> prod(range(1, 3))
    2

    >>> prod(range(1, 3), 2.)
    4.
    """
    iterable = list(iterable)
    return reduce(mul, iterable, *arg)


def r2_score(y_true, y_pred, axis=None):
    y_true = np.asanyarray(y_true)
    y_pred = np.asanyarray(y_pred)

    mse = np.square(y_true - y_pred).mean(axis=axis)
    var = y_true.var(axis=axis)

    return 1 - mse / var


def propagate_uncertainty(cov: np.ndarray, jac: np.ndarray) -> np.ndarray:
    """Propagates the uncertainty of a covariance matrix given the
    passed Jacobian

    If the variable arrays are multidimensional then will output in
    the shape of the arrays

    see https://en.wikipedia.org/wiki/Propagation_of_uncertainty
    """
    cov = np.asanyarray(cov)

    var_ndim = cov.ndim // 2
    det_ndim = jac.ndim - var_ndim
    det_shape, var_shape = jac.shape[:det_ndim], jac.shape[det_ndim:]
    assert var_shape == cov.shape[:var_ndim] == cov.shape[var_ndim:]

    var_size = np.prod(var_shape, dtype=int)
    det_size = np.prod(det_shape, dtype=int)

    cov2d = cov.reshape((var_size, var_size))
    jac2d = jac.reshape((det_size, var_size))

    det_cov2d = np.linalg.multi_dot((jac2d, cov2d, jac2d.T))
    det_cov = det_cov2d.reshape(det_shape + det_shape)
    return det_cov


def rescale_to_artists(artists, ax=None):
    import matplotlib.pyplot as plt

    ax = ax or plt.gca()
    while True:
        r = ax.figure.canvas.get_renderer()
        extents = [
            t.get_window_extent(renderer=r).transformed(ax.transData.inverted())
            for t in artists
        ]
        min_extent = np.min([e.min for e in extents], axis=0)
        max_extent = np.max([e.max for e in extents], axis=0)
        min_lim, max_lim = zip(ax.get_xlim(), ax.get_ylim())

        # Sometimes the window doesn't always rescale first time around
        if (min_extent < min_lim).any() or (max_extent > max_lim).any():
            extent = max_extent - min_extent
            max_extent += extent * 0.05
            min_extent -= extent * 0.05
            xlim, ylim = zip(
                np.minimum(min_lim, min_extent), np.maximum(max_lim, max_extent)
            )
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)
        else:
            break

    return xlim, ylim


# These may no longer be needed?
def add_arrays(*arrays: np.ndarray) -> np.ndarray:
    """Sums over broadcasting multidimensional arrays
    whilst preserving the total sum

    a = np.arange(10).reshape(1, 2, 1, 5)
    b = np.arange(8).reshape(2, 2, 2, 1)

    >>> add_arrays(a, b).sum()
    73.0
    >>> add_arrays(a, b).shape
    (2, 2, 2, 5)
    >>> a.sum() + b.sum()
    73
    """
    b = np.broadcast(*arrays)
    return sum(a * np.size(a) / b.size for a in arrays)


Axis = Optional[Union[bool, int, Tuple[int, ...]]]


def aggregate(array: np.ndarray, axis: Axis = None, **kwargs) -> np.ndarray:
    """
    aggregates the values of array

    if axis is False then aggregate returns the unmodified array

    otherwise aggrate returns np.sum(array, axis=axis, **kwargs)
    """
    if axis is False:
        return array

    return np.sum(array, axis=axis, **kwargs)


def diag(array: np.ndarray, *ds: Tuple[int, ...]) -> np.ndarray:
    array = np.asanyarray(array)
    d1 = array.shape
    if ds:
        ds = (d1,) + ds
    else:
        ds = (d1, d1)

    out = np.zeros(sum(ds, ()))
    diag_inds = tuple(map(np.ravel, (i for d in ds for i in np.indices(d))))
    out[diag_inds] = array.ravel()
    return out



import cProfile
import logging
import multiprocessing as mp
import os
from os import path
from typing import Iterable

from dynesty.dynesty import _function_wrapper
from emcee.ensemble import _FunctionWrapper

from autoconf import conf
from autofit.non_linear.paths.abstract import AbstractPaths
from .process import AbstractJob, Process, StopCommand

logger = logging.getLogger(
    __name__
)


def _is_likelihood_function(
        function
) -> bool:
    """
    Is the function a callable used to evaluate likelihood?

    Naturally in Autofit this would be a child of the Fitness class.
    In Dynesty the likelihood function is wrapped in _function_wrapper
    and called 'loglikelihood'

    Parameters
    ----------
    function
        Some object

    Returns
    -------
    Is the object a log likelihood function?
    """
    from autofit.non_linear.abstract_search import NonLinearSearch
    return any([
        isinstance(
            function,
            NonLinearSearch.Fitness
        ),
        isinstance(
            function,
            _function_wrapper
        ) and function.name == 'loglikelihood',
        isinstance(
            function,
            _FunctionWrapper
        )
    ])


class SneakyJob(AbstractJob):
    def __init__(self, function, *args):
        """
        A job performed on a process.

        If the function is the log likelihood function then it is set to None.

        If the log likelihood function is in the args, it is filtered from the args,
        but its index retained.

        This prevents large amounts of data comprised in an Analysis class from being
        copied over to processes multiple times.

        Parameters
        ----------
        function
            Some function to which a pool.map has been applied
        args
            The arguments to that function
        """
        super().__init__()
        if _is_likelihood_function(function):
            self.function = None
        else:
            self.function = function

        self.args = list()
        self.fitness_index = None

        for i, arg in enumerate(args):
            if _is_likelihood_function(
                    arg
            ):
                if self.fitness_index is not None:
                    raise AssertionError(
                        f"Two arguments of type NonLinear.Fitness passed to {function.__name__}"
                    )
                self.fitness_index = i
            else:
                self.args.append(arg)

    @property
    def is_not_fitness(self) -> bool:
        """
        If the map is not being applied to a fitness function and a fitness
        function is not one of the arguments then this map need not be sneaky
        """
        return self.function is not None and self.fitness_index is None

    def perform(self, likelihood_function):
        """
        Computes the log likelihood. The likelihood function
        is passed from a copy associated with the current process.

        Depending on whether the likelihood function itself is
        being mapped, or some function mapped onto the likelihood
        function as an argument, the likelihood function will be
        called or added to the arguments.

        Parameters
        ----------
        likelihood_function
            A likelihood function associated with the processes
            to avoid copying data for every single function call

        Returns
        -------
        The log likelihood
        """
        if self.function is None:
            return likelihood_function(
                self.args
            )
        if self.is_not_fitness:
            return self.function(self.args)
        args = (
                self.args[:self.fitness_index]
                + [likelihood_function]
                + self.args[self.fitness_index:]
        )
        return self.function(
            args
        )


class SneakyProcess(Process):
    def __init__(
            self,
            name: str,
            paths: AbstractPaths,
            initializer=None,
            initargs=None,
            job_args=tuple()
    ):
        """
        Each SneakyProcess creates its own queue to avoid locking during
        highly parallel optimisations.
        """

        super().__init__(
            name,
            job_queue=mp.Queue(),
            initializer=initializer,
            initargs=initargs,
            job_args=job_args,
        )
        self.paths = paths

    def run(self):
        """
        Run this process, completing each job in the job_queue and
        passing the result to the queue.

        The process continues to execute until a StopCommand is passed.
        This occurs when the SneakyMap goes out of scope.
        """

        if conf.instance["general"]["profiling"]["parallel_profile"]:
            pr = cProfile.Profile()
            pr.enable()

        self._init()
        self.logger.debug("starting")
        while True:
            job = self.job_queue.get()
            if job is StopCommand:
                self.logger.debug(
                    "StopCommand found"
                )
                break
            try:
                self.queue.put(
                    job.perform(
                        *self.job_args
                    )
                )
            except Exception as e:
                self.logger.exception(e)
                self.queue.put(e)
        self.logger.debug("terminating process {}".format(self.name))
        self.job_queue.close()

        if conf.instance["general"]["test"]["parallel_profile"]:

            try:
                os.makedirs(self.paths.profile_path)
            except FileExistsError:
                pass

            sneaky_path = path.join(self.paths.profile_path, f"sneaky_{self.pid}.prof")

            pr.dump_stats(sneaky_path)
            pr.disable()

    def open_profiler(self):

        if conf.instance["general"]["test"]["parallel_profile"]:
            pr = cProfile.Profile()
            pr.enable()


class SneakyPool:
    def __init__(
            self,
            processes: int,
            fitness,
            paths: AbstractPaths,
            initializer=None,
            initargs=None
    ):
        """
        Implements the same interface as multiprocessing's pool,
        but associates the fitness object with each process to
        prevent data being copies to each process for every function
        call.

        Parameters
        ----------
        processes
            The number of cores to be used simultaneously.
        fitness
            A class comprising data and a model which can be used
            to evaluate the likelihood of a live point
        initializer
        initargs
        """
        logger.debug(
            f"Creating SneakyPool with {processes} processes"
        )
        self._processes = [
            SneakyProcess(
                str(number),
                paths=paths,
                initializer=initializer,
                initargs=initargs,
                job_args=(fitness,)
            )
            for number in range(processes)
        ]
        for process in self.processes:
            process.start()

    @property
    def processes(self):
        return self._processes

    def map(self, function, args_list):
        """
        Execute the function with the given arguments across all of the
        processes. The likelihood  argument is removed from each args in
        the args_list.

        Parameters
        ----------
        function
            Some function
        args_list
            An iterable of iterables of arguments passed to the function

        Yields
        ------
        Results from the function evaluation
        """
        jobs = [
            SneakyJob(
                function,
                *(
                    (args,) if not isinstance(
                        args,
                        Iterable
                    ) else tuple(args)
                )
            ) for args in args_list
        ]

        logger.debug(
            f"Running {len(jobs)} jobs across {self.processes} processes"
        )

        for i, job in enumerate(jobs):
            process = self.processes[
                i % len(self.processes)
                ]
            process.job_queue.put(
                job
            )

        target = len(jobs)
        count = 0

        exception = None

        while count < target:
            for process in self.processes:
                if not process.queue.empty():
                    item = process.queue.get()
                    count += 1
                    if isinstance(
                            item,
                            Exception
                    ):
                        exception = item
                    else:
                        yield item

        logger.debug(
            "All jobs complete"
        )

        if exception is not None:
            raise exception

    def __del__(self):
        """
        Called when the map goes out of scope.

        Tell each process to terminate with a StopCommand and then join
        each process with a timeout of one second.
        """
        logger.debug(
            "Deconstructing SneakyMap"
        )

        logger.debug(
            "Terminating processes..."
        )
        for process in self.processes:
            process.job_queue.put(StopCommand)

        logger.debug(
            "Joining processes..."
        )
        for process in self.processes:
            try:
                process.join(0.5)
            except Exception as e:
                logger.exception(e)
